/* $Id$ */
/**************************************************************************
 *
 *  FILE            giftransp.c
 *  MODULE OF       
 *
 *  DESCRIPTION     
 *
 *  WRITTEN BY      Sverre H. Huseby
 *
 *  MODIFICATIONS   
 *
 **************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <shhmsg.h>
#include <xalloc.h>

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

typedef unsigned char BYTE;
typedef unsigned short WORD;

static FILE *in, *out;
static BYTE buff[256];


/**************************************************************************
 *                                                                        *
 *                        P U B L I C    D A T A                          *
 *                                                                        *
 **************************************************************************/



/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

static void readn(void *buff, int n)
{
    if (fread(buff, 1, n, in) != n)
	msgFatal("read error when trying to read %d bytes\n", n);
}

static void writen(void *buff, int n)
{
    if (fwrite(buff, 1, n, out) != n)
	msgFatal("write error when trying to write %d bytes\n", n);
}

static void openFiles(const char *filename)
{
    char *bakname;
    struct stat st;

    bakname = (char *) xmalloc(strlen(filename) + 2);
    strcpy(bakname, filename);
    strcat(bakname, "~");
    if (rename(filename, bakname) < 0)
	msgFatalPerror("couldn't rename %s to %s", filename, bakname);
    if ((in = fopen(bakname, "rb")) == NULL)
	msgFatal("couldn't read inputfile\n");
    if ((out = fopen(filename, "wb")) == NULL)
	msgFatal("couldn't write outputfile\n");
    stat(bakname, &st);
    chmod(filename, st.st_mode);
    free(bakname);
}

static void skipSubBlocks(void)
{
    int n;
	
    for (;;) {
	readn(buff, 1);  /* size */
	writen(buff, 1);
	n = buff[0];
	msgVerbose(4, "copying subblock, size %d bytes\n", n);
	if (n == 0)
	    break;
	readn(buff, n);
	writen(buff, n);
    }
}

static void setTransparencyIndex(int tridx)
{
    int gceseen = 0; /* seen a graphic control extension? */
    
    /*
     *  copy file header and logical screen descriptor
     */
    msgVerbose(2, "copying header and logical screen descriptor\n");
    readn(buff, 6 + 7);
    /* fix version. at least 89a is needed for transparency */
    if (buff[3] == '8' && buff[4] < '9') {
	buff[4] = '9';
	msgVerbose(1, "converting to GIF89a\n");
    }
    writen(buff, 6 + 7);

    /*
     *  possibly copy the global color table
     */
    if (buff[10] & 0x80) {
	int gctsiz = buff[10] & 7;

	msgVerbose(2, "copying global color table\n");
	readn(buff, 3 * (1 << (gctsiz + 1)));
	writen(buff, 3 * (1 << (gctsiz + 1)));
    }

    while (!feof(in)) {
	/*
	 *  read the label of the next block to see what we should do.
	 */
	readn(buff, 1);
	if (buff[0] == 0x2C) {
	    /* image descriptor */
	    if (!gceseen) {
		static BYTE gce[8] = {
		    0x21,   /* extension introducer */
		    0xF9,   /* graphic control label */
		    0x04,   /* block size */
		    0x01,   /* flags: transparent index given */
		    0, 0,   /* delay time; not given */
		    0,      /* transparent color index */
		    0       /* block terminator */
		};
		
		msgVerbose(1, "adding graphic control extension "
			   "with transparency\n");
		gce[6] = tridx;
		writen(gce, sizeof(gce));
	    }
	    msgVerbose(2, "copying image descriptor\n");
	    readn(buff + 1, 9);
	    writen(buff, 10);
	    /*
	     *  possibly copy the local color table
	     */
	    if (buff[9] & 0x80) {
		int lctsiz = 1 << ((buff[9] & 7) + 1);

		msgVerbose(2, "copying local color table, %d entries\n",
			   lctsiz);
		readn(buff, 3 * lctsiz);
		writen(buff, 3 * lctsiz);
	    }
	    /* image data follows. first lzw minimum code size */
	    msgVerbose(2, "copying compressed image data\n");
	    readn(buff, 1);
	    writen(buff, 1);
	    /* now the image data subblocks */
	    skipSubBlocks();
	    gceseen = 0;
	} else if (buff[0] == 0x21) {
	    /* extension */
	    readn(buff + 1, 1);
	    writen(buff, 2);
	    if (buff[1] == 0xF9) {
		msgVerbose(2, "copying graphic control extension; "
			   "setting up transparency\n");
		readn(buff, 6);
		buff[1] |= 1;
		buff[4] = tridx;
		writen(buff, 6);
		gceseen = 1;
	    } else {
		msgVerbose(2, "copying extension block, type 0x%02X\n",
			   (int) buff[1]);
		skipSubBlocks();
	    }
	} else if (buff[0] == 0x3B) {
	    /* stream trailer */
	    msgVerbose(2, "copying trailer\n");
	    writen(buff, 1);
	    break;
	}
    }
}


/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

int main(int argc, char *argv[])
{
    msgSetName(argv[0]);
    msgSetVerbose(10);
    if (argc != 3) {
	printf("usage: %s gif-file transparent-index\n", msgGetName());
	return 0;
    }
    openFiles(argv[1]);
    setTransparencyIndex(atoi(argv[2]));
    return 0;
}
